/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.40 from the
 * contents of GCrypt.xs. Do not edit this file, edit GCrypt.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "GCrypt.xs"
/*
 ===========================================================================
 Crypt::GCrypt

 Perl interface to the GNU Cryptographic library
 
 Author: Alessandro Ranellucci <aar@cpan.org>
 
 Use this software AT YOUR OWN RISK.
 ===========================================================================
*/

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "ppport.h"

#include <gcrypt.h>
#include <string.h>

#ifdef USE_ITHREADS
    #ifdef I_PTHREAD
        #include <pthread.h>
        #define HAVE_PTHREAD
    #else
        #warning "Perl ithreads not available or not implemented with Pthread: building a non-threadsafe Crypt::GCrypt"
    # endif
#endif


static const char my_name[] = "Crypt::GCrypt";
static const char author[] = "Alessandro Ranellucci <aar@cpan.org>";

enum cg_type
{
    CG_TYPE_CIPHER,
    CG_TYPE_ASYMM,
    CG_TYPE_DIGEST
};
enum cg_action
{
    CG_ACTION_NONE,
    CG_ACTION_ENCRYPT,
    CG_ACTION_DECRYPT
};
enum cg_padding
{
    CG_PADDING_NONE,
    CG_PADDING_STANDARD,
    CG_PADDING_NULL,
    CG_PADDING_SPACE
};

struct Crypt_GCrypt_s {
    int type;
    int action;
    gcry_cipher_hd_t h;
/*
    gcry_ac_handle_t h_ac;
*/
    gcry_md_hd_t h_md;
/*
    gcry_ac_key_t key_ac;
*/
    gcry_error_t err;
    int mode;
    int padding;
    unsigned char *buffer;
    STRLEN buflen, blklen, keylen;
    int need_to_call_finish;
    int buffer_is_decrypted;
};
typedef struct Crypt_GCrypt_s *Crypt_GCrypt;

/* a Crypt_GCrypt_MPI need only be a pointer to a gcrypt MPI object */
typedef gcry_mpi_t Crypt_GCrypt_MPI;

/* return the offset of padding or -1 if none */
int find_padding (Crypt_GCrypt gcr, unsigned char *string, size_t string_len) {
    unsigned char last_char = string[string_len-1];
    size_t i, offset;
    void *p;
    
    switch (gcr->padding) {
        case CG_PADDING_STANDARD:
            /* padding length is last_char */
            for (i = 1; i <= last_char; ++i) {
                if (string[string_len-i] != last_char) return -1;
            }
            return string_len-last_char;
            
        case CG_PADDING_NULL:
            p = memchr((char *) string, '\0', string_len);
            if (p == NULL) return -1;
            
            offset = p - (void *)string;
            for (i = offset; i < string_len; ++i) {
                if (string[string_len-i] != '\0') return -1;
            }
            return offset;
            
        case CG_PADDING_SPACE:
            p = memchr((char *) string, '\32', string_len);
            if (p == NULL) return -1;
            
            offset = p - (void *)string;
            for (i = offset; i < string_len; ++i) {
                if (string[string_len-i] != '\32') return -1;
            }
            return offset;
    }
    return -1;
}

Crypt_GCrypt_MPI
dereference_gcm(SV* sv_gcm) {
    if (!sv_derived_from(sv_gcm, "Crypt::GCrypt::MPI"))
        croak("Not a Crypt::GCrypt::MPI object");
    IV tmp = SvIV((SV*)SvRV(sv_gcm));
    return INT2PTR(Crypt_GCrypt_MPI, tmp);
}

#ifdef HAVE_PTHREAD
    GCRY_THREAD_OPTION_PTHREAD_IMPL;
#endif

void
init_library() {
  gcry_error_t ret;
  if (gcry_control(GCRYCTL_INITIALIZATION_FINISHED_P)) {
    /* we just need to make sure that the right version is available */
    if (!gcry_check_version(GCRYPT_VERSION))
      croak("libgcrypt version mismatch (needed: %s)", GCRYPT_VERSION);
    return;
  }
  /* else, we need to go ahead with the full initialization: */
  #ifdef HAVE_PTHREAD
    ret = gcry_control(GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread);
    if (gcry_err_code(ret) != GPG_ERR_NO_ERROR)
    croak("could not initialize libgcrypt for threads (%d: %s/%s)", 
      gcry_err_code(ret),
      gcry_strsource(ret),
      gcry_strerror(ret));
  #endif
  
  if (!gcry_check_version(GCRYPT_VERSION))
    croak("libgcrypt version mismatch (needed: %s)", GCRYPT_VERSION);

  /* Why do it this way?  see 
     /usr/share/doc/libgcrypt11-doc/html/Initializing-the-library.html#sample-use-suspend-secmem
  */

  /* We don't want to see any warnings, e.g. because we have not yet
     parsed program options which might be used to suppress such
     warnings. */
  gcry_control (GCRYCTL_SUSPEND_SECMEM_WARN);
     
  /* Allocate a pool of 32k secure memory.  This make the secure memory
     available and also drops privileges where needed.  

     This mirrors changes made in libgcrypt 1.4.3, to auto-initialize
     the library with 32KB of secure memory if no other initialization
     has been done.

     FIXME: we should probably allow the user to choose how much 
     secure RAM to use something like this:

     use Crypt::GCrypt { secmem => 1024*1024 };

  */
  gcry_control (GCRYCTL_INIT_SECMEM, 32768, 0);
     
  /* It is now okay to let Libgcrypt complain when there was/is
     a problem with the secure memory. */
  gcry_control (GCRYCTL_RESUME_SECMEM_WARN);

  gcry_control(GCRYCTL_INITIALIZATION_FINISHED);
}


#line 192 "GCrypt.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

#line 336 "GCrypt.c"

XS_EUPXS(XS_Crypt__GCrypt_built_against_version); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_built_against_version)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
	SV *	RETVAL;
#line 187 "GCrypt.xs"
        init_library();
        RETVAL = newSVpvn(GCRYPT_VERSION, strlen(GCRYPT_VERSION));
#line 349 "GCrypt.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_gcrypt_version); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_gcrypt_version)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    {
	SV *	RETVAL;
#line 195 "GCrypt.xs"
        const char * v;
#line 367 "GCrypt.c"
#line 197 "GCrypt.xs"
        init_library();
        v = gcry_check_version(NULL);
        RETVAL = newSVpvn(v, strlen(v));
#line 372 "GCrypt.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_new); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_new)
{
    dVAR; dXSARGS;
    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
    {
	Crypt_GCrypt	RETVAL;
#line 207 "GCrypt.xs"
        char *s, *algo_s = NULL, *mode_s, *key_s;
        int i, algo, mode;
        unsigned int c_flags, ac_flags, md_flags;
/*
        gcry_ac_id_t ac_algo;
*/
        bool have_mode;
#line 396 "GCrypt.c"
#line 215 "GCrypt.xs"
        New(0, RETVAL, 1, struct Crypt_GCrypt_s);
        s = SvPV_nolen(ST(0));
        if (strcmp(s, "Crypt::GCrypt") == 0) {
            i = 1;
        } else {
            i = 0;
        }
        if ((items-i % 2) == 1)
            croak("Wrong number of arguments for Crypt::GCrypt->new()");

        /* Default values: */
        RETVAL->type = -1;
        RETVAL->padding = CG_PADDING_STANDARD;
        RETVAL->action = CG_ACTION_NONE;
        RETVAL->need_to_call_finish = 0;
        RETVAL->buffer_is_decrypted = 0;
        RETVAL->buffer = NULL;
        c_flags = 0;
        ac_flags = 0;
        md_flags = 0;
        have_mode = 0;

        /* Let's get parameters: */
        while (i < items) {
            s = SvPV_nolen(ST(i));
            if (strcmp(s, "type") == 0) {
                s = SvPV_nolen(ST(i+1));
                if (strcmp(s, "cipher") == 0)
                    RETVAL->type = CG_TYPE_CIPHER;
                if (strcmp(s, "asymm") == 0)
                    RETVAL->type = CG_TYPE_ASYMM;
                if (strcmp(s, "digest") == 0)
                    RETVAL->type = CG_TYPE_DIGEST;
            }
            if (strcmp(s, "algorithm") == 0) {
                algo_s = SvPV_nolen(ST(i+1));
            }
            if (strcmp(s, "mode") == 0) {
                mode_s = SvPV_nolen(ST(i+1));
                have_mode = 1;
            }
            if (strcmp(s, "padding") == 0) {
                s = SvPV_nolen(ST(i+1));
                if (strcmp(s, "none") == 0)
                    RETVAL->padding = CG_PADDING_NONE;
                if (strcmp(s, "standard") == 0)
                    RETVAL->padding = CG_PADDING_STANDARD;
                if (strcmp(s, "null") == 0)
                    RETVAL->padding = CG_PADDING_NULL;
            }
            if (strcmp(s, "secure") == 0) {
                if (SvTRUE(ST(i+1))) {
                   c_flags |= GCRY_CIPHER_SECURE;
                   md_flags |= GCRY_MD_FLAG_SECURE;
                }
            }
            if (strcmp(s, "hmac") == 0) {
                key_s = SvPV(ST(i+1), RETVAL->keylen);
                md_flags |= GCRY_MD_FLAG_HMAC;
            }
            if (strcmp(s, "enable_sync") == 0) {
                if (SvTRUE(ST(i+1))) c_flags |= GCRY_CIPHER_ENABLE_SYNC;
            }
            i = i + 2;
        }
        if (RETVAL->type == -1)
            croak("No valid type specified for Crypt::GCrypt->new()");
        if (!algo_s)
            croak("No algorithm specified for Crypt::GCrypt->new()");

        init_library();

        if (RETVAL->type == CG_TYPE_CIPHER) {
            /* Checking algorithm */
            if (!(algo = gcry_cipher_map_name(algo_s)))
                croak("Unknown cipher algorithm %s", algo_s);
            RETVAL->blklen = gcry_cipher_get_algo_blklen(algo);
            RETVAL->keylen = gcry_cipher_get_algo_keylen(algo);

            /* Checking mode */
            if (have_mode) {
                switch (mode_s[0]) {
                    case 'e':
                        if (strcmp(mode_s+1, "cb") == 0)
                            RETVAL->mode = GCRY_CIPHER_MODE_ECB;
                        break;
                    case 'c':
                        if (strcmp(mode_s+1, "fb") == 0)
                            RETVAL->mode = GCRY_CIPHER_MODE_CFB;
                        else if (strcmp(mode_s+1, "bc") == 0)
                            RETVAL->mode = GCRY_CIPHER_MODE_CBC;
                        break;
                    case 's':
                        if (strcmp(mode_s+1, "tream") == 0)
                            RETVAL->mode = GCRY_CIPHER_MODE_STREAM;
                        break;
                    case 'o':
                        if (strcmp(mode_s+1, "fb") == 0)
                            RETVAL->mode = GCRY_CIPHER_MODE_OFB;
                        break;
                }
            } else {
                RETVAL->mode = RETVAL->blklen > 1 ? GCRY_CIPHER_MODE_CBC
                        : GCRY_CIPHER_MODE_STREAM;
            }
            if (!RETVAL->mode)
                croak("Unknown mode %s", mode_s);

            /* Init cipher */
            RETVAL->err = gcry_cipher_open(&RETVAL->h, algo, RETVAL->mode, c_flags);
            if (RETVAL->h == NULL) XSRETURN_UNDEF;
        }
        if (RETVAL->type == CG_TYPE_DIGEST) {
            if (!(algo = gcry_md_map_name(algo_s)))
                croak("Unknown digest algorithm %s", algo_s);

            RETVAL->err = gcry_md_open(&RETVAL->h_md, algo, md_flags);
            if (RETVAL->h_md == NULL) XSRETURN_UNDEF;

            if (md_flags & GCRY_MD_FLAG_HMAC) {
                /* what if this overwrites the earlier error value? */
                RETVAL->err = gcry_md_setkey(RETVAL->h_md, key_s, RETVAL->keylen);
            }
        }
        if (RETVAL->type == CG_TYPE_ASYMM) {

            croak("Asymmetric cryptography is not yet supported by Crypt::GCrypt");
	    XSRETURN_UNDEF;

        }


#line 530 "GCrypt.c"
	{
	    SV * RETVALSV;
	    RETVALSV = sv_newmortal();
	    sv_setref_pv(RETVALSV, "Crypt::GCrypt",
		    (void *)RETVAL);
	    ST(0) = RETVALSV;
	}
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_encrypt); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_encrypt)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "gcr, in");
    {
	Crypt_GCrypt	gcr;
	SV *	in = ST(1)
;
#line 355 "GCrypt.xs"
        char *ibuf, *curbuf, *obuf;
        size_t len, ilen, buflen;
#line 556 "GCrypt.c"
	SV *	RETVAL;

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 358 "GCrypt.xs"
        if (gcr->action != CG_ACTION_ENCRYPT)
            croak("start('encrypting') was not called");

        ibuf = SvPV(in, ilen);

        if (gcr->padding == CG_PADDING_NONE && ilen % gcr->blklen > 0)
            croak("'None' padding requires that input to ->encrypt() is supplied as a multiple of blklen");

        /* Get total buffer+ibuf length */
        Newz(0, curbuf, ilen + gcr->buflen, char);
        memcpy(curbuf, gcr->buffer, gcr->buflen);
        memcpy(curbuf+gcr->buflen, ibuf, ilen);

        if ((len = (ilen+gcr->buflen) % gcr->blklen) == 0) {
            len = ilen+gcr->buflen;
            gcr->buffer[0] = '\0';
            gcr->buflen = 0;
        } else {
            char *tmpbuf;
            len = (ilen+gcr->buflen) - len;   /* len contiene i byte da scrivere effettivemente */

            Newz(0, tmpbuf, len, char);
            memcpy(tmpbuf, curbuf, len);
            memcpy(gcr->buffer, curbuf+len, (ilen+gcr->buflen)-len);
            gcr->buflen = (ilen+gcr->buflen)-len;
            Safefree(curbuf);
            curbuf = tmpbuf;
        }

        /* Encrypt data */
        New(0, obuf, len, char);
        if (len > 0) {
            if ((gcr->err = gcry_cipher_encrypt(gcr->h, obuf, len, curbuf, len)) != 0)
                croak("encrypt: %s", gcry_strerror(gcr->err));
        }
        RETVAL = newSVpvn(obuf, len);
        Safefree(curbuf);
        Safefree(obuf);
#line 604 "GCrypt.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_finish); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_finish)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gcr");
    {
	Crypt_GCrypt	gcr;
#line 403 "GCrypt.xs"
        char *obuf;
        size_t rlen, return_len, padding_length;
#line 623 "GCrypt.c"
	SV *	RETVAL;

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 406 "GCrypt.xs"
        if (gcr->type != CG_TYPE_CIPHER)
            croak("Can't call finish when doing non-cipher operations");
        gcr->need_to_call_finish = 0;
        if (gcr->action == CG_ACTION_ENCRYPT) {

            if (gcr->buflen < gcr->blklen) {
                unsigned char *tmpbuf;
                rlen = gcr->blklen - gcr->buflen;
                Newz(0, tmpbuf, gcr->buflen+rlen, unsigned char);
                memcpy(tmpbuf, gcr->buffer, gcr->buflen);
                switch (gcr->padding) {
                    case CG_PADDING_STANDARD:
                        memset(tmpbuf + gcr->buflen, rlen, rlen);
                        break;
                    case CG_PADDING_NULL:
                        memset(tmpbuf + gcr->buflen, 0, rlen);
                        break;
                    case CG_PADDING_SPACE:
                        memset(tmpbuf + gcr->buflen, '\32', rlen);
                        break;
                }
                Safefree(gcr->buffer);
                gcr->buffer = tmpbuf;
            } else {
                if (gcr->padding == CG_PADDING_NULL && gcr->blklen == 8) {
                    unsigned char *tmpbuf;
                    Newz(0, tmpbuf, gcr->buflen+8, unsigned char);
                    memcpy(tmpbuf, gcr->buffer, gcr->buflen);
                    memset(tmpbuf + gcr->buflen, 0, 8);
                    Safefree(gcr->buffer);
                    gcr->buffer = tmpbuf;
                }
            }
            Newz(0, obuf, gcr->blklen, char);
            if ((gcr->err = gcry_cipher_encrypt(gcr->h, obuf, gcr->blklen, gcr->buffer, gcr->blklen)) != 0)
                croak("encrypt: %s", gcry_strerror(gcr->err));
            gcr->buffer[0] = '\0';
            gcr->buflen = 0;
            RETVAL = newSVpvn(obuf, gcr->blklen);
            Safefree(obuf);

        } else {  /* CG_ACTION_DECRYPT */

            /* decrypt remaining ciphertext if any */
            New(0, obuf, gcr->buflen, char);
            return_len = gcr->buflen;
            if (gcr->buflen > 0) {
                if (gcr->buffer_is_decrypted == 1) {
                    Move(gcr->buffer, obuf, gcr->buflen, char);
                } else {
                    if ((gcr->err = gcry_cipher_decrypt(gcr->h, obuf, return_len, gcr->buffer, gcr->buflen)) != 0)
                        croak("decrypt: %s", gcry_strerror(gcr->err));
                }
                gcr->buffer[0] = '\0';
                gcr->buflen = 0;

                /* Remove padding */
                return_len = find_padding(gcr, (unsigned char *) obuf, return_len);
            }

            RETVAL = newSVpvn(obuf, return_len);
            Safefree(obuf);
        }
#line 696 "GCrypt.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_decrypt); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_decrypt)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "gcr, in");
    {
	Crypt_GCrypt	gcr;
	SV *	in = ST(1)
;
#line 479 "GCrypt.xs"
        unsigned char *ibuf, *obuf, *ciphertext, *decrypted_buffer;
        size_t total_len, len, ilen;
        int ciphertext_offset;
#line 718 "GCrypt.c"
	SV *	RETVAL;

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 483 "GCrypt.xs"
        if (gcr->action != CG_ACTION_DECRYPT)
            croak("start('decrypting') was not called");

        ibuf = (unsigned char *) SvPV(in, ilen);
        if ((ilen % gcr->blklen) > 0 || ilen == 0)
            croak("input must be a multiple of blklen");

        /* Concatenate buffer and input to get total length of ciphertext */
        total_len = gcr->buflen + ilen;  /* total_len is a multiple of blklen */
        Newz(0, ciphertext, total_len, unsigned char);
        Move(gcr->buffer, ciphertext, gcr->buflen, unsigned char);
        Move(ibuf, ciphertext+gcr->buflen, ilen, unsigned char);

        /* if our buffer was decrypted by the previous run of this method,
           we set a ciphertext_offset to avoid re-decrypting such plaintext
           coming from the buffer */
        ciphertext_offset = (gcr->buffer_is_decrypted == 1) ? gcr->buflen : 0;

        /* strip last block and move it to buffer */
        len = total_len - gcr->blklen;  /* len is the length of plaintext we're returning */
        Move(ciphertext+len, gcr->buffer, (total_len - len), unsigned char);
        gcr->buflen = gcr->blklen;

        /* do actual decryption */
        New(0, obuf, len, unsigned char);
        Copy(ciphertext, obuf, ciphertext_offset, unsigned char);
        if (len-ciphertext_offset > 0) { /* that is, if we have something to decrypt */
            if ((gcr->err = gcry_cipher_decrypt(gcr->h, obuf+ciphertext_offset, len-ciphertext_offset, ciphertext+ciphertext_offset, len-ciphertext_offset)) != 0)
                croak("decrypt: %s", gcry_strerror(gcr->err));
        }
        Safefree(ciphertext);

        /* OPTIMIZATION for compatibility with implementations of Crypt::GCrypt <= 1.17:
           decrypt buffer and check if it seems padded */
        if ((gcr->err = gcry_cipher_decrypt(gcr->h, gcr->buffer, gcr->buflen, NULL, 0)) != 0) /* in-place decryption */
                         croak("decrypt: %s", gcry_strerror(gcr->err));
        gcr->buffer_is_decrypted = 1;
        if (find_padding(gcr, gcr->buffer, gcr->buflen) == -1) {
            /* if the string doesn't appear to be padded, let's append it to the
               output so that users who don't call ->finish() don't break their applications */
            Renew(obuf, len + gcr->buflen, unsigned char);
            Move(gcr->buffer, obuf+len, gcr->buflen, unsigned char);
            len = len + gcr->buflen;
            gcr->buffer[0] = '\0';
            gcr->buflen = 0;
            gcr->buffer_is_decrypted = 0;
        }
        RETVAL = newSVpvn((char *) obuf, len);
        Safefree(obuf);
#line 777 "GCrypt.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_start); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_start)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "gcr, act");
    {
	Crypt_GCrypt	gcr;
	SV *	act = ST(1)
;
#line 540 "GCrypt.xs"
        char *action;
        size_t len;
#line 798 "GCrypt.c"

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 543 "GCrypt.xs"
        gcr->err = gcry_cipher_reset(gcr->h);
        Safefree(gcr->buffer);
        New(0, gcr->buffer, gcr->blklen, unsigned char);
        gcr->buflen = 0;
        gcr->need_to_call_finish = 1;
        action = SvPV(act, len);
        switch (action[0]) {
            case 'e':
                gcr->action = CG_ACTION_ENCRYPT;
                break;
            case 'd':
                gcr->action = CG_ACTION_DECRYPT;
                break;
        }
#line 821 "GCrypt.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Crypt__GCrypt_setkey); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_setkey)
{
    dVAR; dXSARGS;
    if (items < 1)
       croak_xs_usage(cv,  "gcr, ...");
    {
	Crypt_GCrypt	gcr;
#line 562 "GCrypt.xs"
        char *k, *s;
        char *mykey, *buf;
/*
        gcry_ac_key_type_t keytype;
        gcry_ac_data_t keydata;
*/
        gcry_mpi_t mpi;
        size_t len;
#line 844 "GCrypt.c"

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 571 "GCrypt.xs"
        /* Set key for cipher */
        if (gcr->type == CG_TYPE_CIPHER) {
            buf = NULL;
            mykey = SvPV(ST(1), len);
            /* If key is shorter than our algorithm's key size 
               let's pad it with zeroes */
            if (len < gcr->keylen) {
                Newz(0, buf, gcr->keylen, char);
                memcpy(buf, mykey, len);
                mykey = buf;
            }
            gcr->err = gcry_cipher_setkey(gcr->h, mykey, gcr->keylen);
            if (gcr->err != 0) croak("setkey: %s", gcry_strerror(gcr->err));
            Safefree(buf);
        }

        /* Set key for asymmetric criptography */
        if (gcr->type == CG_TYPE_ASYMM) {
            croak("Asymmetric cryptography is not yet supported by Crypt::GCrypt");
        }
#line 873 "GCrypt.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Crypt__GCrypt_setiv); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_setiv)
{
    dVAR; dXSARGS;
    if (items < 1)
       croak_xs_usage(cv,  "gcr, ...");
    {
	Crypt_GCrypt	gcr;
#line 596 "GCrypt.xs"
        char *buf, *param;
        size_t len;
#line 890 "GCrypt.c"

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 599 "GCrypt.xs"
        buf = NULL;
        if (gcr->type != CG_TYPE_CIPHER)
            croak("Can't call setiv when doing non-cipher operations");
        if (items == 2) {
            param = SvPV(ST(1), len);
            if (len < gcr->blklen) {
                Newz(0, buf, gcr->blklen, char);
                memcpy(buf, param, len);
                param = buf;
            }
        } else if (items == 1) {
            Newz(0, buf, gcr->blklen, char);
            param = buf;
        } else
            croak("Usage: $cipher->setiv([iv])");
        gcry_cipher_setiv(gcr->h, param, gcr->blklen);
        Safefree(buf);
#line 916 "GCrypt.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Crypt__GCrypt_sync); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_sync)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gcr");
    {
	Crypt_GCrypt	gcr;

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 621 "GCrypt.xs"
        if (gcr->type != CG_TYPE_CIPHER)
            croak("Can't call sync when doing non-cipher operations");
        gcry_cipher_sync(gcr->h);
#line 941 "GCrypt.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Crypt__GCrypt_keylen); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_keylen)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gcr");
    {
	Crypt_GCrypt	gcr;
	int	RETVAL;
	dXSTARG;

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 629 "GCrypt.xs"
        if (gcr->type != CG_TYPE_CIPHER)
            croak("Can't call keylen when doing non-cipher operations");
        RETVAL = gcr->keylen;
#line 968 "GCrypt.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_blklen); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_blklen)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gcr");
    {
	Crypt_GCrypt	gcr;
	int	RETVAL;
	dXSTARG;

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 639 "GCrypt.xs"
        if (gcr->type != CG_TYPE_CIPHER)
            croak("Can't call blklen when doing non-cipher operations");
        RETVAL = gcr->blklen;
#line 996 "GCrypt.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_reset); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_reset)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gcr");
    {
	Crypt_GCrypt	gcr;

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 649 "GCrypt.xs"
        if (gcr->type != CG_TYPE_DIGEST)
            croak("Can't call reset when doing non-digest operations");
        gcry_md_reset(gcr->h_md);
#line 1022 "GCrypt.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Crypt__GCrypt_write); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_write)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "gcr, in");
    {
	Crypt_GCrypt	gcr;
	SV *	in = ST(1)
;
#line 658 "GCrypt.xs"
        char *ibuf;
        size_t ilen;
#line 1041 "GCrypt.c"

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 661 "GCrypt.xs"
        if (gcr->type != CG_TYPE_DIGEST)
            croak("Can't call write when doing non-digest operations.");
        ibuf = SvPV(in, ilen);
        gcry_md_write(gcr->h_md, ibuf, ilen);
#line 1054 "GCrypt.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Crypt__GCrypt_read); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_read)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gcr");
    {
	Crypt_GCrypt	gcr;
#line 670 "GCrypt.xs"
        unsigned char *output;
        size_t len;
#line 1071 "GCrypt.c"
	SV *	RETVAL;

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 673 "GCrypt.xs"
        if (gcr->type != CG_TYPE_DIGEST)
            croak("Can't call read when doing non-digest operations.");
        output = gcry_md_read(gcr->h_md, 0);
        len = gcry_md_get_algo_dlen(gcry_md_get_algo(gcr->h_md));
        RETVAL = newSVpvn((const char *) output, len);
#line 1086 "GCrypt.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_digest_length); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_digest_length)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gcr");
    {
	Crypt_GCrypt	gcr;
	int	RETVAL;
	dXSTARG;

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 685 "GCrypt.xs"
        if (gcr->type != CG_TYPE_DIGEST)
            croak("Can't call digest_length when doing non-digest operations");
        RETVAL = gcry_md_get_algo_dlen(gcry_md_get_algo(gcr->h_md));
#line 1115 "GCrypt.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_clone); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_clone)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gcr");
    {
	Crypt_GCrypt	gcr;
	Crypt_GCrypt	RETVAL;

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 695 "GCrypt.xs"
        if (gcr->type != CG_TYPE_DIGEST)
            croak("Crypt::GCrypt::clone() is only currently defined for digest objects");

        New(0, RETVAL, 1, struct Crypt_GCrypt_s);
        Copy(gcr, RETVAL, 1, struct Crypt_GCrypt_s);
        /* if we allow clone() for cipher objects, we should duplicate the buffer */
        RETVAL->err = gcry_md_copy(&RETVAL->h_md, gcr->h_md);
        if (RETVAL->h_md == NULL) XSRETURN_UNDEF;
#line 1147 "GCrypt.c"
	{
	    SV * RETVALSV;
	    RETVALSV = sv_newmortal();
	    sv_setref_pv(RETVALSV, "Crypt::GCrypt",
		    (void *)RETVAL);
	    ST(0) = RETVALSV;
	}
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_digest_algo_available); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_digest_algo_available)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "algo");
    {
	SV *	algo = ST(0)
;
#line 710 "GCrypt.xs"
        const char *algo_s;
        int algo_id;
#line 1172 "GCrypt.c"
	int	RETVAL;
	dXSTARG;
#line 713 "GCrypt.xs"
        algo_s = SvPV_nolen(algo);
        init_library();
        algo_id = gcry_md_map_name(algo_s);
        if (algo_id) {
            if (gcry_md_test_algo(algo_id))
                RETVAL = 0;
            else
                RETVAL = 1;
        } else {
            RETVAL = 0;
        }
#line 1187 "GCrypt.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_cipher_algo_available); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_cipher_algo_available)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "algo");
    {
	SV *	algo = ST(0)
;
#line 731 "GCrypt.xs"
        const char *algo_s;
        int algo_id;
#line 1206 "GCrypt.c"
	int	RETVAL;
	dXSTARG;
#line 734 "GCrypt.xs"
        algo_s = SvPV_nolen(algo);
        init_library();
        algo_id = gcry_cipher_map_name(algo_s);
        if (algo_id) {
            if (gcry_cipher_algo_info(algo_id, GCRYCTL_TEST_ALGO, 0, 0))
                RETVAL = 0;
            else
                RETVAL = 1;
        } else {
            RETVAL = 0;
        }
#line 1221 "GCrypt.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt_DESTROY); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt_DESTROY)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gcr");
    {
	Crypt_GCrypt	gcr;

	if (sv_derived_from(ST(0), "Crypt::GCrypt")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcr = (Crypt_GCrypt) tmp;
	} else
		croak("gcr is not of type Crypt::GCrypt")
;
#line 752 "GCrypt.xs"
        if (gcr->type == CG_TYPE_CIPHER) gcry_cipher_close(gcr->h);
/*
        if (gcr->type == CG_TYPE_ASYMM)  gcry_ac_close(gcr->h_ac);
*/
        if (gcr->type == CG_TYPE_DIGEST) gcry_md_close(gcr->h_md);

        if (gcr->need_to_call_finish == 1)
            warn("WARNING: the ->finish() method was not called after encryption/decryption.");

        Safefree(gcr->buffer);
        Safefree(gcr);
#line 1255 "GCrypt.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_new); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_new)
{
    dVAR; dXSARGS;
    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
    {
#line 783 "GCrypt.xs"
        char *s;
        int i, valix, secure, set, format;
        Crypt_GCrypt_MPI src;
        size_t len;
        gcry_error_t err;       
#line 1274 "GCrypt.c"
	Crypt_GCrypt_MPI	RETVAL;
#line 789 "GCrypt.xs"
        RETVAL = NULL;
        secure = 0;
        valix = -1;
        format = GCRYMPI_FMT_STD;
        set = 0;
        s = SvPV_nolen(ST(0));
        if (strcmp(s, "Crypt::GCrypt::MPI") == 0) {
            i = 1;
        } else {
            i = 0;
        }
        if ((items-i > 1) && ((items-i % 2) == 1))
            croak("Wrong number of arguments for Crypt::GCrypt::MPI->new()");
        if (items-i == 1) {
           /* this is the value */
           valix = i;
        } else {
           /* this is a parameterized list */
           while (i < items) {
             s = SvPV_nolen(ST(i));

             if (strcmp(s, "secure") == 0)
                if (SvTRUE(ST(i+1)))
                  secure = 1;
             if (strcmp(s, "format") == 0)
                format = SvIV(ST((i+1)));
             if (strcmp(s, "value") == 0)
                valix = i + 1;
             i += 2;
           }
        }

        /* we're copying an mpi: */
        if ((valix >= 0) && sv_derived_from(ST(valix), "Crypt::GCrypt::MPI")) {
          IV tmp = SvIV((SV*)SvRV(ST(valix))); /* found this incantation at */
          src = INT2PTR(Crypt_GCrypt_MPI,tmp); /* http://cpansearch.perl.org/src/DSTH/Math-GSL-Linalg-SVD-0.0.2/SVD.c */
          if (secure && ! gcry_mpi_get_flag(src, GCRYMPI_FLAG_SECURE)) {
            /* we were asked for a secure MPI, but we were given an insecure one to copy */
            RETVAL=gcry_mpi_snew(gcry_mpi_get_nbits(src));
            if (RETVAL == NULL) XSRETURN_UNDEF;
            gcry_mpi_set(RETVAL, src);
          } else {
            /* this will give us a secure MPI if a secure MPI was passed in, even if we
               also got secure => 0 ; a bit weird, but on the safe side. */
            RETVAL=gcry_mpi_copy(src);
          }
        } else {
          if (secure) {
            RETVAL=gcry_mpi_snew(0);
          } else {
            RETVAL=gcry_mpi_new(0);
          }
          if (RETVAL == NULL) XSRETURN_UNDEF;

          /* do something with the info at valix */
          if (valix >= 0) {
            switch(SvTYPE(ST(valix))) {
              case SVt_PVIV:
                /* if negative value, we have to jump through some hoops: */
                if (SvIV(ST(valix)) < 0) {
                    RETVAL=gcry_mpi_set_ui(NULL, 0);
                    gcry_mpi_sub_ui(RETVAL, RETVAL, -1*(SvIV(ST(valix))));
                } else {
                /* this can be dealt with by regular unsigned ints */
                    gcry_mpi_set_ui(RETVAL, SvIV(ST(valix)));
                }
                break;
              case SVt_PV:
                /* handle bytestrings 
                   */
                s = SvPV(ST(valix), len);
                src = NULL;
                err = gcry_mpi_scan(&src, format, s, ((format == GCRYMPI_FMT_HEX) ? 0 : len), NULL);
                if (err != 0) croak("Crypt::GCrypt::MPI::new (from string, with format %d) libgcrypt internal failure %s", format, gcry_strerror(err));
                /* FIXME: can we "eat" the head of the the input string if format is FMT_PGP or FMT_SSH ? */
                /* avoid memory leak: */
                if (secure) {
                   gcry_mpi_set(RETVAL, src);
                   gcry_mpi_release(src);
                } else {
                   gcry_mpi_release(RETVAL);
                   RETVAL = src;
                }
                break;
              default:
                croak("value argument for Crypt::GCrypt::MPI->new() must currently be either an int or another Crypt::GCrypt::MPI (%d, %d, %d)", SvTYPE(ST(valix)), valix, format);
                break;
            }
          }
        }
#line 1367 "GCrypt.c"
	{
	    SV * RETVALSV;
	    RETVALSV = sv_newmortal();
	    sv_setref_pv(RETVALSV, "Crypt::GCrypt::MPI",
		    (void *)RETVAL);
	    ST(0) = RETVALSV;
	}
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_swap); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_swap)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "sv_gcma, gcmb");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;
#line 887 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_swap(gcma, gcmb);
        ST(0) = sv_gcma;
        XSRETURN(1);
#line 1404 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_set); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_set)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "sv_gcma, gcmb");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;
#line 897 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_set(gcma, gcmb);
        ST(0) = sv_gcma;
        XSRETURN(1);
#line 1435 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_is_secure); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_is_secure)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gcm");
    {
	Crypt_GCrypt_MPI	gcm;
	bool	RETVAL;

	if (sv_derived_from(ST(0), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcm = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcm is not of type Crypt::GCrypt::MPI")
;
#line 906 "GCrypt.xs"
        RETVAL=gcry_mpi_get_flag(gcm, GCRYMPI_FLAG_SECURE);
#line 1460 "GCrypt.c"
	ST(0) = boolSV(RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_cmp); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_cmp)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "gcma, gcmb");
    {
	Crypt_GCrypt_MPI	gcma;
	Crypt_GCrypt_MPI	gcmb;
	int	RETVAL;
	dXSTARG;

	if (sv_derived_from(ST(0), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcma = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcma is not of type Crypt::GCrypt::MPI")
;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;
#line 915 "GCrypt.xs"
        RETVAL=gcry_mpi_cmp(gcma, gcmb);
#line 1494 "GCrypt.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_copy); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_copy)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gcm");
    {
	Crypt_GCrypt_MPI	gcm;
	Crypt_GCrypt_MPI	RETVAL;

	if (sv_derived_from(ST(0), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcm = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcm is not of type Crypt::GCrypt::MPI")
;
#line 923 "GCrypt.xs"
        RETVAL=gcry_mpi_copy(gcm);
#line 1519 "GCrypt.c"
	{
	    SV * RETVALSV;
	    RETVALSV = sv_newmortal();
	    sv_setref_pv(RETVALSV, "Crypt::GCrypt::MPI",
		    (void *)RETVAL);
	    ST(0) = RETVALSV;
	}
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_add); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_add)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "sv_gcma, gcmb");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;
#line 932 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_add(gcma, gcma, gcmb);
        ST(0) = sv_gcma;
        XSRETURN(1);
#line 1556 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_addm); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_addm)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "sv_gcma, gcmb, gcmm");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;
	Crypt_GCrypt_MPI	gcmm;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;

	if (sv_derived_from(ST(2), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(2)));
		gcmm = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmm is not of type Crypt::GCrypt::MPI")
;
#line 943 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_addm(gcma, gcma, gcmb, gcmm);
        ST(0) = sv_gcma;
        XSRETURN(1);
#line 1595 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_sub); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_sub)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "sv_gcma, gcmb");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;
#line 953 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_sub(gcma, gcma, gcmb);
        ST(0) = sv_gcma;
        XSRETURN(1);
#line 1626 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_subm); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_subm)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "sv_gcma, gcmb, gcmm");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;
	Crypt_GCrypt_MPI	gcmm;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;

	if (sv_derived_from(ST(2), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(2)));
		gcmm = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmm is not of type Crypt::GCrypt::MPI")
;
#line 964 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_subm(gcma, gcma, gcmb, gcmm);
        ST(0) = sv_gcma;
        XSRETURN(1);
#line 1665 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_mul); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_mul)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "sv_gcma, gcmb");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;
#line 974 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_mul(gcma, gcma, gcmb);
        ST(0) = sv_gcma;
        XSRETURN(1);
#line 1696 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_mulm); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_mulm)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "sv_gcma, gcmb, gcmm");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;
	Crypt_GCrypt_MPI	gcmm;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;

	if (sv_derived_from(ST(2), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(2)));
		gcmm = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmm is not of type Crypt::GCrypt::MPI")
;
#line 985 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_mulm(gcma, gcma, gcmb, gcmm);
        ST(0) = sv_gcma;
        XSRETURN(1);
#line 1735 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_mul_2exp); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_mul_2exp)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "sv_gcm, e");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcm = ST(0)
;
	int	e = (int)SvIV(ST(1))
;
#line 995 "GCrypt.xs"
        Crypt_GCrypt_MPI gcm = dereference_gcm(sv_gcm);
        if (e >= 0) {
          /* this can be dealt with by regular unsigned ints */
          gcry_mpi_mul_2exp(gcm, gcm, e);
        } else {
          croak("exponent argument for Crypt::GCrypt::MPI::mul_2exp() must be an unsigned integer");
        }
        ST(0) = sv_gcm;
        XSRETURN(1);
#line 1765 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_div); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_div)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "sv_gcma, gcmb");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;
#line 1010 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_div(gcma, NULL, gcma, gcmb, 0);
        ST(0) = sv_gcma;
        /* FIXME: should we return the remainder as well, if called in a list context? */
        XSRETURN(1);
#line 1797 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_mod); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_mod)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "sv_gcma, gcmb");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;
#line 1021 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_mod(gcma, gcma, gcmb);
        ST(0) = sv_gcma;
        XSRETURN(1);
#line 1828 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_powm); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_powm)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "sv_gcma, gcme, gcmm");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcme;
	Crypt_GCrypt_MPI	gcmm;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcme = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcme is not of type Crypt::GCrypt::MPI")
;

	if (sv_derived_from(ST(2), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(2)));
		gcmm = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmm is not of type Crypt::GCrypt::MPI")
;
#line 1032 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_powm(gcma, gcma, gcme, gcmm);
        ST(0) = sv_gcma;
        XSRETURN(1);
#line 1867 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_invm); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_invm)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "sv_gcma, gcmb");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;
#line 1042 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_invm(gcma, gcma, gcmb);
        ST(0) = sv_gcma;
        /* FIXME: should we do anything with the return value (1 if invm actually exists)? */
        XSRETURN(1);
#line 1899 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_gcd); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_gcd)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "sv_gcma, gcmb");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcma = ST(0)
;
	Crypt_GCrypt_MPI	gcmb;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;
#line 1053 "GCrypt.xs"
        Crypt_GCrypt_MPI gcma = dereference_gcm(sv_gcma);
        gcry_mpi_gcd(gcma, gcma, gcmb);
        ST(0) = sv_gcma;
        XSRETURN(1);
#line 1930 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_mutually_prime); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_mutually_prime)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "gcma, gcmb");
    {
	Crypt_GCrypt_MPI	gcma;
	Crypt_GCrypt_MPI	gcmb;
#line 1063 "GCrypt.xs"
        Crypt_GCrypt_MPI gcd;
#line 1948 "GCrypt.c"
	bool	RETVAL;

	if (sv_derived_from(ST(0), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcma = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcma is not of type Crypt::GCrypt::MPI")
;

	if (sv_derived_from(ST(1), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(1)));
		gcmb = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcmb is not of type Crypt::GCrypt::MPI")
;
#line 1065 "GCrypt.xs"
        gcd = gcry_mpi_new(0);
        RETVAL=gcry_mpi_gcd(gcd, gcma, gcmb);
        gcry_mpi_release(gcd);
#line 1968 "GCrypt.c"
	ST(0) = boolSV(RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_dump); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_dump)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv_gcm");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV*	sv_gcm = ST(0)
;
#line 1075 "GCrypt.xs"
        Crypt_GCrypt_MPI gcm = dereference_gcm(sv_gcm);
        gcry_mpi_dump(gcm);
        ST(0) = sv_gcm;
        XSRETURN(1);
#line 1991 "GCrypt.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_print); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_print)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "gcm, format");
    {
	Crypt_GCrypt_MPI	gcm;
	int	format = (int)SvIV(ST(1))
;
#line 1085 "GCrypt.xs"
        size_t size;
        unsigned char* buf;
        gcry_error_t err;
#line 2012 "GCrypt.c"
	SV *	RETVAL;

	if (sv_derived_from(ST(0), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gcm = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gcm is not of type Crypt::GCrypt::MPI")
;
#line 1089 "GCrypt.xs"
        /* FIXME: make format default to FMT_STD somehow (how do we not require a parameter?) */
        err = gcry_mpi_print(format, NULL, 0, &size,  gcm);
        if (err != 0) croak("GCrypt::MPI::print start: %s", gcry_strerror(err));
        /* FMT_HEX asks for an extra byte for the null char, but perl allocates that
           already, so we treat that case special */
        RETVAL = newSVpv("", ((format == GCRYMPI_FMT_HEX) ? size-1 : size));
        buf = (unsigned char*) SvPV_nolen(RETVAL);
        err = gcry_mpi_print(format, buf, size, &size,  gcm);
        if (err != 0) croak("GCrypt::MPI::print finish: %s", gcry_strerror(err));
#line 2031 "GCrypt.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Crypt__GCrypt__MPI_DESTROY); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Crypt__GCrypt__MPI_DESTROY)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "gmpi");
    {
	Crypt_GCrypt_MPI	gmpi;

	if (sv_derived_from(ST(0), "Crypt::GCrypt::MPI")) {
		IV tmp = SvIV((SV*)SvRV(ST(0)));
		gmpi = (Crypt_GCrypt_MPI) tmp;
	} else
		croak("gmpi is not of type Crypt::GCrypt::MPI")
;
#line 1105 "GCrypt.xs"
        gcry_mpi_release(gmpi);
        gmpi = NULL;
#line 2057 "GCrypt.c"
    }
    XSRETURN_EMPTY;
}

#ifdef __cplusplus
extern "C"
#endif
XS_EXTERNAL(boot_Crypt__GCrypt); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_Crypt__GCrypt)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSXSAPIVERCHK;
#endif
#if (PERL_REVISION == 5 && PERL_VERSION < 9)
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5)
    XS_VERSION_BOOTCHECK;
#  ifdef XS_APIVERSION_BOOTCHECK
    XS_APIVERSION_BOOTCHECK;
#  endif
#endif

        newXS_deffile("Crypt::GCrypt::built_against_version", XS_Crypt__GCrypt_built_against_version);
        newXS_deffile("Crypt::GCrypt::gcrypt_version", XS_Crypt__GCrypt_gcrypt_version);
        (void)newXSproto_portable("Crypt::GCrypt::new", XS_Crypt__GCrypt_new, file, "@");
        newXS_deffile("Crypt::GCrypt::encrypt", XS_Crypt__GCrypt_encrypt);
        newXS_deffile("Crypt::GCrypt::finish", XS_Crypt__GCrypt_finish);
        newXS_deffile("Crypt::GCrypt::decrypt", XS_Crypt__GCrypt_decrypt);
        newXS_deffile("Crypt::GCrypt::start", XS_Crypt__GCrypt_start);
        newXS_deffile("Crypt::GCrypt::setkey", XS_Crypt__GCrypt_setkey);
        newXS_deffile("Crypt::GCrypt::setiv", XS_Crypt__GCrypt_setiv);
        newXS_deffile("Crypt::GCrypt::sync", XS_Crypt__GCrypt_sync);
        newXS_deffile("Crypt::GCrypt::keylen", XS_Crypt__GCrypt_keylen);
        newXS_deffile("Crypt::GCrypt::blklen", XS_Crypt__GCrypt_blklen);
        newXS_deffile("Crypt::GCrypt::reset", XS_Crypt__GCrypt_reset);
        newXS_deffile("Crypt::GCrypt::write", XS_Crypt__GCrypt_write);
        newXS_deffile("Crypt::GCrypt::read", XS_Crypt__GCrypt_read);
        newXS_deffile("Crypt::GCrypt::digest_length", XS_Crypt__GCrypt_digest_length);
        newXS_deffile("Crypt::GCrypt::clone", XS_Crypt__GCrypt_clone);
        newXS_deffile("Crypt::GCrypt::digest_algo_available", XS_Crypt__GCrypt_digest_algo_available);
        newXS_deffile("Crypt::GCrypt::cipher_algo_available", XS_Crypt__GCrypt_cipher_algo_available);
        newXS_deffile("Crypt::GCrypt::DESTROY", XS_Crypt__GCrypt_DESTROY);
        (void)newXSproto_portable("Crypt::GCrypt::MPI::new", XS_Crypt__GCrypt__MPI_new, file, "@");
        newXS_deffile("Crypt::GCrypt::MPI::swap", XS_Crypt__GCrypt__MPI_swap);
        newXS_deffile("Crypt::GCrypt::MPI::set", XS_Crypt__GCrypt__MPI_set);
        newXS_deffile("Crypt::GCrypt::MPI::is_secure", XS_Crypt__GCrypt__MPI_is_secure);
        newXS_deffile("Crypt::GCrypt::MPI::cmp", XS_Crypt__GCrypt__MPI_cmp);
        newXS_deffile("Crypt::GCrypt::MPI::copy", XS_Crypt__GCrypt__MPI_copy);
        newXS_deffile("Crypt::GCrypt::MPI::add", XS_Crypt__GCrypt__MPI_add);
        newXS_deffile("Crypt::GCrypt::MPI::addm", XS_Crypt__GCrypt__MPI_addm);
        newXS_deffile("Crypt::GCrypt::MPI::sub", XS_Crypt__GCrypt__MPI_sub);
        newXS_deffile("Crypt::GCrypt::MPI::subm", XS_Crypt__GCrypt__MPI_subm);
        newXS_deffile("Crypt::GCrypt::MPI::mul", XS_Crypt__GCrypt__MPI_mul);
        newXS_deffile("Crypt::GCrypt::MPI::mulm", XS_Crypt__GCrypt__MPI_mulm);
        newXS_deffile("Crypt::GCrypt::MPI::mul_2exp", XS_Crypt__GCrypt__MPI_mul_2exp);
        newXS_deffile("Crypt::GCrypt::MPI::div", XS_Crypt__GCrypt__MPI_div);
        newXS_deffile("Crypt::GCrypt::MPI::mod", XS_Crypt__GCrypt__MPI_mod);
        newXS_deffile("Crypt::GCrypt::MPI::powm", XS_Crypt__GCrypt__MPI_powm);
        newXS_deffile("Crypt::GCrypt::MPI::invm", XS_Crypt__GCrypt__MPI_invm);
        newXS_deffile("Crypt::GCrypt::MPI::gcd", XS_Crypt__GCrypt__MPI_gcd);
        newXS_deffile("Crypt::GCrypt::MPI::mutually_prime", XS_Crypt__GCrypt__MPI_mutually_prime);
        newXS_deffile("Crypt::GCrypt::MPI::dump", XS_Crypt__GCrypt__MPI_dump);
        newXS_deffile("Crypt::GCrypt::MPI::print", XS_Crypt__GCrypt__MPI_print);
        newXS_deffile("Crypt::GCrypt::MPI::DESTROY", XS_Crypt__GCrypt__MPI_DESTROY);

    /* Initialisation Section */

#line 767 "GCrypt.xs"
    { /* found this method of storing constants in
         http://blogs.sun.com/akolb/entry/pitfals_of_the_perl_xs */
    HV *stash;

    stash = gv_stashpv("Crypt::GCrypt::MPI", TRUE);
    newCONSTSUB(stash, "FMT_STD", newSViv(GCRYMPI_FMT_STD));
    newCONSTSUB(stash, "FMT_PGP", newSViv(GCRYMPI_FMT_PGP));
    newCONSTSUB(stash, "FMT_SSH", newSViv(GCRYMPI_FMT_SSH));
    newCONSTSUB(stash, "FMT_HEX", newSViv(GCRYMPI_FMT_HEX));
    newCONSTSUB(stash, "FMT_USG", newSViv(GCRYMPI_FMT_USG));
    }

#line 2148 "GCrypt.c"

    /* End of Initialisation Section */

#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

